#include "hash2.hpp"
#include <map>
#include <algorithm>
#include <sstream>
#include <fcgi_stdio.h>

static string kEmptyStr;

#define Info ""

const string sm_ContentTypeName    = "Content-Type";
const string sm_ContentTypeDefault = "text/html";



const string& CONFIG_DICT_PATH("ac/");
#undef LOG_
#define LOG_(...)


/////////////////////////////////////////////////////////////////////////////
//  Run test (printout arguments obtained from command-line)

std::map<string,CPmHash2Isam*> dict_set;

void SetHeader(std::ostream& c_out, const string& header, const string& val)
{
	c_out << header << ": " << val << endl;
}


void SetContentType(std::ostream& c_out, const string& contentType)
{
	SetHeader(c_out, sm_ContentTypeName , contentType);
}



static string find_sugg(CPmHash2Isam *cur_dict, const string& in, char *out, int out_sz, bool single_attempt)
{
	string prefix;

	for(prefix = in;
		!cur_dict->Find(prefix.c_str(), out, out_sz, NULL) && !prefix.empty();
		prefix = in.substr(0, prefix.length() - 1))
	{
		std::cerr << "find_sugg: Unsuccessfully search: Term=" << in << "; SG=" << prefix << endl;
		if(single_attempt)
			return kEmptyStr;
	}

	std::cerr << "find_sugg: Found: Term=" << in << "; SG=" << prefix << endl;

	return prefix;
}

unsigned decode16(char c) {
   unsigned rc = c - '0';
   return rc < 10? rc : (c | ' ') - 'a' + 10;
}


string UrlDecode(const char* urlEncoded)
{
	int encodedLen = strlen(urlEncoded);
	string result;

	for (int i = 0; i < encodedLen; i++)
	{
		char c = urlEncoded[i];
		if (c == '+')
			c = ' ';
		else if (c == '%')
		{
			if (i + 2 >= encodedLen)
				c = 0;
			else
				c = (decode16(urlEncoded[i+1]) << 4) | decode16(urlEncoded[i+2]);
			i += 2;
		}

		result += c;
	}

	return result;
}


string GetRequestValue(const char *str, const string& key)
{
	string query_string(str);

	int keyIdx = 0;
	do
	{
		keyIdx = query_string.find(key, keyIdx);

		// to handle the "subkey" situation, i.e.
		// when query string is 'abcdef=1&cdef=2&ef=3' and we are looking for 'ef' key
		if (keyIdx == 0 || query_string[keyIdx-1] == '&')
			break;

		// key was not found
		if (keyIdx == string::npos)
			return "";
	} while (keyIdx > 0);

	keyIdx += key.size();

	int valIdx = query_string.find("&", keyIdx);
	if (valIdx == string::npos)
		return query_string.substr(keyIdx);
	else
		return query_string.substr(keyIdx, valIdx - keyIdx);
}

/*----------------------------------------------------------------------------------------*/

void ProcessAutoCompeteRequest(const char* query_string, int print_cerr)
{
    LOG_("Query_String: " << query_string);

    string dict_name(UrlDecode(GetRequestValue(query_string, "dict=").c_str()));
    string term(UrlDecode(GetRequestValue(query_string, "q=").c_str()));
    string single_attempt(UrlDecode(GetRequestValue(query_string, "single_attempt=").c_str()));

    LOG_("dict: " << dict_name);
    LOG_("term: " << term);
    LOG_("single_attempt: " << single_attempt);

    CPmHash2Isam* cur_dict;

    if(!dict_set.count(dict_name))
	{
		cur_dict = new CPmHash2Isam;
		string dict_path(CONFIG_DICT_PATH + dict_name);

		if(!cur_dict->Open(dict_path))
		{
			delete cur_dict;
			cur_dict = NULL;
			std::cerr << "Can't open dictionary file: " << dict_path << endl;
		}

		dict_set[dict_name] = cur_dict;
		std::cerr << "Dictionary file cached: " << dict_path << endl;
	}
	else
	{
		cur_dict = dict_set[dict_name];
		std::cerr << "Dictionary cache accepted: " << dict_name << endl;
    }

	//std::ostrstream output_content(g_OutputBuffer, sizeof(g_OutputBuffer));



	std::ostringstream c_out;

    //CNcbiOstream& c_out = output_content; // std::cout;
//    CNcbiOstream& c_out = std::cout;

    SetContentType(c_out, "text/js; charset=UTF-8");
    c_out << endl;	// in HTTP content and headers are delimited by 2 consecutive \r\n

    if(cur_dict == NULL)
	{
		c_out <<"NSuggest_CreateData(\"_dictionary_error\", new Array(\"\"), 4);\n"; // 4 - invalid db
		return;
    }

    char bufhh[100000];
    bufhh[0] = 0;

    string prefix = find_sugg(cur_dict, term, bufhh, sizeof(bufhh), !single_attempt.empty());

    if(bufhh[0] && !prefix.empty())  // If returned something
    {
	if(bufhh[0] & 4)	// Gzipped data
	{
		std::cerr << "Don't know how to work with GZIPed data!" << endl;
		return;

/*
	        char buf_zip[100000];
			size_t src_read, dst_written;

			if(BASE64_Decode(bufhh + 2, 100000, &src_read, buf_zip, 100000, &dst_written))
			{
				string accept_encoding(ctx.GetRequest().GetRandomProperty("ACCEPT_ENCODING"));
				std:cerr << "INFO: HTTP_ACCEPT_ENCODING=" + accept_encoding << endl;
				if(accept_encoding.find("gzip") != string::npos) // can send gzipped
				{
					response.SetHeaderValue("Content-Encoding","gzip");
					response.WriteHeader();
					c_out.write(buf_zip, dst_written);
				}
				else  // need to unzip
				{
					response.WriteHeader();
					CZipCompression unzipper;
					unzipper.SetFlags(CZipCompression::fCheckFileHeader);
 					if(unzipper.DecompressBuffer(buf_zip, 100000, bufhh, 100000, &dst_written))
					{
						bufhh[dst_written] = 0;
						c_out <<  bufhh;
					}
					else
					{
						c_out << "NSuggest_CreateData(\"" << prefix << "\", new Array(\"UNZIP!\"), 3);\n"; // 2 - empty result
					}
				}
			}
*/

	}
	else			// Ordinary uncompressed data
	{
		if(bufhh[0] & 040)  // Edit for Nsuggest
		{
			int flags = 0;
			if((bufhh[0] & 1) == 0) flags |= 1; // Complete result
			if((bufhh[0] & 2) != 0) flags |= 8; // Match from begin
               		if(flags == 0) flags = 64; // not null for activation


			string result(bufhh + 2);
			const string new_delimiter("\", \"");
			const string old_delimiter("|");

			string::size_type start_pos = 0;

			for (int i = 0; ; i++)
			{
				start_pos = result.find(old_delimiter, start_pos);
				if (start_pos == string::npos)
					break;

				result.replace(start_pos, old_delimiter.size(), new_delimiter);
				start_pos += new_delimiter.size();
			}

				c_out << "NSuggest_CreateData(\"" << prefix << "\", new Array(\""
					<< result << "\"), " << flags << ");\n";

//				c_out << "NSuggest_CreateData(\"" << prefix << "\", new Array(\""
//					<< NStr::Replace(bufhh + 2, "|", "\", \"") << "\"), "<< flags << ");\n";
		}
		else
		{
			c_out << &bufhh[2] << "\n";
		}
	}
   }
   else
   {
 	c_out << "NSuggest_CreateData(\"" << term << "\", new Array(\"\"), 3);\n"; // 2 - empty result + 1 - complete result
   }

   LOG_("sending to printf: " << endl << c_out.str());
   printf(c_out.str().c_str());

   if(print_cerr) 
     std::cerr << "RC=[" << c_out.str() << "]" << endl;

  return;
}


/*----------------------------------------------------------------------------------------*/

/////////////////////////////////////////////////////////////////////////////
//  MAIN

int main(int argc, const char* argv[])
{
  if(argc == 2)
    ProcessAutoCompeteRequest(argv[1], 1);
  else
    while (FCGI_Accept() >= 0)
	ProcessAutoCompeteRequest(getenv("QUERY_STRING"), 0);

  return 0;
}
