#include <sys/stat.h>
#include <fcntl.h>
#include "hash2.hpp"
#include <algorithm>

#define STR_LEN 200000

/******************************************************************/

static int htab_vals[] = { 
	312931, 45645, 890798, 90, 659876, 8695869, 900124, 543653453,
	75943,  53453, 98054, 234, 867431, 1091436, 552546, 56445211	
 };

const char PM_TERM_SEPARATOR = 0x10;	// CTRL+P

 
class hash 
{
public:
	hash(const char *p) 
	{
		hash1 = 0x5f5f5f55 ^ p[0] ^ (-p[1] << 8);
		hash2 = 0x3cf0f51f ^ p[1] ^ (-p[0] << 16);

	    while(char c = *p++) 
		{
			hash1 = (hash1 << 7) ^ (hash1 >> (32-7)) ^ (htab_vals[(hash1 ^ c) & 0xF] + c);
			// hash1 = (hash1 << 1) ^ (htab_vals[hash1 & 0xF] + c) ^ (hash1 >> 22);
			hash2 = (hash2 << 2) ^ (htab_vals[hash2 & 0xF] - c) ^ (hash2 >> 30);
		}

		passport = hash2 & 0x1F;
		hash2  >>= 5;// leave 27 bits
	}
  
	Uint4 hash1;		// Primary hash (used for index in hashtable)
	Uint4 hash2;		// Secondary hash (used for linear search comparing)
	Uint2 passport;		// Current string passport
};


typedef struct // Internal - uses only for create new indexes 
{
	Uint4 hash1;   // Primary hash
	Uint4 offset;  // Offset in the text term file
	unsigned int hash2:27;  // Secondary hash 
	unsigned int str_len:5; // Ceiled binary logarifm of real str_len
	Uint4 term_no;          // Term number
	Uint2 passport; // Current string passport
} TERM_ENTRY_CREAT;


// Compare class for sorting TERM_ENTRY_CREAT structs
struct TERM_ENTRY_CREAT_less 
{
	bool operator()(const TERM_ENTRY_CREAT& x, const TERM_ENTRY_CREAT& y) const 
	{
		return (x.hash1 == y.hash1) ? x.hash2 < y.hash2 : x.hash1 < y.hash1;
	}
};


// Index suffix generators
static string fname_term_entry(const char *fname) 
{
	return string(fname) + ".hash_term_entry";
}

static string fname_backet_entry(const char *fname) 
{
	return string(fname) + ".htable";
}


// !! Static member!
bool CPmHash2Isam::Create(const char* lpIsamName, Int4 backet_sz, Int4 iFlags)
{
	Uint4 htab_mask;
	Int4  htab_sz;
	char buf[STR_LEN];

	FILE* f_data = fopen(lpIsamName, "r");
	if(f_data == NULL) 
		return false;

	vector<TERM_ENTRY_CREAT> te_vec;
	te_vec.reserve(1 << 18);

	int offset  = 0;
	int line_no = 0;

	LOG_POST("Started for termfile: " << lpIsamName);

	while(fgets(buf, STR_LEN, f_data)) // Scan input term file, line by line
	{
		// Leave only _key_ => cut off value or '\n'
		char *p = strchr(buf, PM_TERM_SEPARATOR);
		if(p) 
			*p = 0;
		else 
		{
			p = strchr(buf, '\n');
			if(p) 
				*p = 0;
		}

		hash h(buf);

		TERM_ENTRY_CREAT tec;
		tec.hash1   = h.hash1;
		tec.hash2   = h.hash2;
		tec.offset  = offset;
		tec.passport= h.passport;
		offset      = ftell(f_data);
		int l = offset - tec.offset - 1; // real str_len
		
		// count ceiled binary logarifm of real str_len
		for(tec.str_len = 1; (1 << tec.str_len) < l; tec.str_len++);

		tec.term_no = line_no; 
		line_no++;

		te_vec.push_back(tec);
	} // while - read input file

	fclose(f_data);

	// Calculate hashtab size as 2^N value, ceil of "real" htab_size
	htab_sz = (te_vec.size() / backet_sz) << 1;

	while(htab_sz & (htab_sz - 1)) // while more than one bit_1 
	{
		htab_sz &= ~(htab_sz & -htab_sz);  // cut low bit 
	}

	// There is htab_size is 2^N; Set up htab_mask
	htab_mask = htab_sz - 1;

	LOG_POST("Termfile has been read; Read #" << te_vec.size() << " lines; Htab_sz="<< htab_sz << "; sorting...");
	     
	// Adjust primary hash to htab_size     
	for(vector<TERM_ENTRY_CREAT>::iterator it = te_vec.begin(); it != te_vec.end(); it++) 
	{
		it->hash1 &= htab_mask;
	}
	    
	sort(te_vec.begin(), te_vec.end(), TERM_ENTRY_CREAT_less());  
	LOG_POST("Sort OK");


	FILE *f_htable, *f_term_e;
	f_htable = fopen(fname_backet_entry(lpIsamName).c_str(), "wb");
	f_term_e = fopen(fname_term_entry  (lpIsamName).c_str(), "wb");

	if(f_htable == NULL || f_term_e == NULL) 
	{
		ERR_POST("Can't open index files for write");
		return false;
	}

	BACKET_ENTRY be;

	TERM_STR_ENTRY te;
	memset(&te, ~0, sizeof(te));
	fwrite(&te, sizeof(te), 1, f_term_e); // fictive element -- begin-barried

	vector<TERM_ENTRY_CREAT>::iterator backet_beg = te_vec.begin();
	vector<TERM_ENTRY_CREAT>::iterator backet_end = backet_beg;

	for (int cur_hash = 0; cur_hash < htab_sz; cur_hash++) 
	{
		be.passport =  0; // Init cur element of hash table
		be.offset   = ~0;
		if(cur_hash == backet_beg->hash1) 
		{
			be.offset = ftell(f_term_e) / sizeof(te); 
			while(backet_end < te_vec.end() && backet_beg->hash1 == backet_end->hash1) 
			{
				te.hash2   = backet_end->hash2;
				te.offset  = backet_end->offset;
				te.term_no = backet_end->term_no;
				te.str_len = backet_end->str_len; 
				be.passport |= 1 << backet_end->passport;
				fwrite(&te, sizeof(te), 1, f_term_e);
				backet_end++;
			}

			memset(&te, ~0, sizeof(te));
			fwrite(&te, sizeof(te), 1, f_term_e); // fictive element -- end-barried
			backet_beg = backet_end;
		}

		fwrite(&be, sizeof(be), 1, f_htable);
	}

	be.passport = 0;
	be.offset   = ftell(f_term_e) / sizeof(te);

	fwrite(&be, sizeof(be), 1, f_htable);  
	  
	fclose(f_htable);
	fclose(f_term_e);
	return true; 
}

bool CPmHash2Isam::Open(const char *fname, Int4 flags) 
{
	Close();
    if (!m_backet_entry_fm.Open(fname_backet_entry(fname)))
		return false;
   
    if(flags & HASH2_NO_MMAP2) 
	{
		m_te = NULL;
		if((m_term_entry_fd = open(fname_term_entry(fname).c_str(), O_RDONLY)) < 0)
			return false;
    } 
	else 
	{    
		m_term_entry_fd = -1;
		if ( !m_term_entry_fm.Open(fname_term_entry(fname)))
			return false;

		m_te = (TERM_STR_ENTRY*)m_term_entry_fm.Image();
	}
    
    if((m_termfile = open(fname, O_RDONLY)) < 0)
      return false;
      
    m_be = (BACKET_ENTRY*)m_backet_entry_fm.Image();
    
    m_htab_sz = (m_backet_entry_fm.Size() / sizeof(BACKET_ENTRY)) & -2;
    m_htab_mask = m_htab_sz - 1;
    
    LOG_POST("CPmHash2Isam::Open(" << fname <<"); Size=" << m_htab_sz << "; Mask=" << m_htab_mask);
    return true;
}

bool CPmHash2Isam::Open(const string& fname, Int4 flags) 
{
	return Open(fname.c_str(), flags);
}

void CPmHash2Isam::Close() 
{
	if(m_be) 
	{
		m_backet_entry_fm.Close();
		m_term_entry_fm.Close();
		close(m_termfile);
		if(m_term_entry_fd >= 0) 
			close(m_term_entry_fd);
		m_be = NULL;
	}
}

bool CPmHash2Isam::Find(const char* term, char* out, Int4 out_sz, Int4Ptr p_term_no) const 
{
    hash h(term);
            
    BACKET_ENTRY *be_ptr = m_be + (h.hash1 & m_htab_mask);
    
    if(!((be_ptr->passport >> h.passport) & 1)) { // Check passport
       return false;
    }
    
    int mmap_sz = -1;
    bool rc = false;
    
    TERM_STR_ENTRY entry_buf[1024];
    
    TERM_STR_ENTRY *te_ptr, *te_ptr0;

    if(m_te == NULL) 
	{
		mmap_sz = sizeof(TERM_STR_ENTRY) * (be_ptr[1].offset - be_ptr->offset);
		te_ptr = entry_buf;
		pread(m_term_entry_fd, entry_buf, mmap_sz, sizeof(TERM_STR_ENTRY) *  be_ptr->offset);
	} 
	else 
	{
		te_ptr = m_te + be_ptr->offset;
    }
        
    for(; ; te_ptr++)	 // Scan backet
	{
       char buf[STR_LEN];

       if(te_ptr->hash2 < h.hash2) // Too few hash2 - go to next;
			continue;
       
       if(te_ptr->hash2 > h.hash2 || te_ptr->offset == ~0)	// Too big hash2 - string not found
			break; // return false;	

	   // Hash2 is EQ! Perhaps, this is our line
       int str_len = 1 << te_ptr->str_len;	// Restore str_len (approx)
       
       int real_len = pread(m_termfile, buf, str_len, te_ptr->offset); // Fetch string
       
       if(real_len < 0)
			break;
       
       buf[real_len] = 0;
       
       char *p_end = strchr(buf + (str_len >> 1), '\n'); // Scan for real end of string
       
       if(p_end)
			*p_end = 0;
       
       char *p = strchr(buf, PM_TERM_SEPARATOR); // Scan for separator key/val
       if(p)
			*p = 0; // And split key and val

       if(strcmp(term, buf) == 0)	 // Key is same => string found!
	   {
			if(out && p) // If val requested
				strncpy(out, p + 1, out_sz); // return it

			if(p_term_no) // Return term_no (NOT IMPLEMENTED)
			{
			   // *p_term_no = ~0; // te_ptr->term_no;
				*p_term_no = te_ptr->term_no;
			}
			rc = true; // Successfull exit
			break;
       } 
	   else 
	   {
			LOG_POST("CPmHash2Isam::Find:False read! Term=" << term << "; str=" << buf);
       }
    }

    return rc;
}
