#ifndef __HASH2_HPP__
#define __HASH2_HPP__ HASH2_HPP

#include <stdint.h>

#include <vector>
#include <iostream>
#include <fstream>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>

#include <unistd.h>


using std::vector;
using std::string;
using std::endl;

typedef uint32_t Uint4;
typedef int32_t Int4;
typedef Int4* Int4Ptr;

typedef uint16_t Uint2;

#define LOG_POST(...) (std::cout << "LOG: " << __VA_ARGS__ << endl);
#define ERR_POST(...) (std::cerr << "ERR: " << __VA_ARGS__ << endl);


typedef struct {  // Hashtable containts those structures
	Uint4 passport; // OR-production of "passports" of each term.
	Uint4 offset;   // offset in the TERM_STR_ENTRY file
} BACKET_ENTRY;

typedef struct {  // Entry structure for each term in the test termfile
	Uint4 offset;           // Offset in the text term file
	Uint4 term_no;          // Term number
	unsigned int hash2:27;  // Secondary hash - 27 bits only
	unsigned int str_len:5; // Ceiled binary logarifm of real str_len
} TERM_STR_ENTRY;

class CNcbiFileMap
{
	void* m_MemoryMappedFile; 
	int m_MMFLength;

public:
	CNcbiFileMap() : m_MemoryMappedFile(MAP_FAILED), m_MMFLength(0) {};
	~CNcbiFileMap() { Close(); }

	bool Open(const string& filename)
	{
		int m_DiskFileDesc = open(filename.c_str(), O_RDONLY);
		if (m_DiskFileDesc == -1)
			return false;

		struct stat  st;						// We need the size.
		fstat(m_DiskFileDesc, &st);            // Obtain the size
		m_MMFLength = st.st_size;
		m_MemoryMappedFile = mmap(NULL, m_MMFLength, PROT_READ, MAP_SHARED, m_DiskFileDesc, 0);
		close(m_DiskFileDesc);	// We don't need this now 

		if (m_MemoryMappedFile == MAP_FAILED)
			return false;

		return true;
	}

	bool Close()
	{
		if (m_MemoryMappedFile == MAP_FAILED)
			return true;

		if (-1 == munmap(m_MemoryMappedFile, m_MMFLength))
			return false;

		m_MemoryMappedFile = MAP_FAILED;
		m_MMFLength = 0;
		return true;
	}

	void* Image()
	{
		return m_MemoryMappedFile;
	}

	int Size() const
	{
		return m_MMFLength;
	}
};


#define HASH2_NO_MMAP2 1

class CPmHash2Isam 
{
  public:

    CPmHash2Isam (void) : m_be(NULL), m_term_entry_fd(-1) {;}
    virtual ~CPmHash2Isam (void) { CPmHash2Isam::Close(); }

    virtual  void Close(void);

    static bool Create(const char*, Int4 backet_sz=18, Int4 flags=0);

    bool Open (const char*, Int4 flags=0);
    bool Open (const string&, Int4 flags=0);
    bool Find (const char*,   /* Input: searchable term                     */
               char* = NULL,  /* Into this buffer will be placed term value */
               Int4 = 0,      /* Size of previous buf                       */ 
		       Int4Ptr = NULL /* must to return term_no; not implemented    */
			  ) const;
	       
private:
    Uint4 m_htab_mask; /* Hash table mask; Set up by Open() */
    Int4  m_htab_sz;   /* Hash table size; Set up by Open() */
    
    CNcbiFileMap m_backet_entry_fm; /* Hashtable mmap  */
    CNcbiFileMap m_term_entry_fm;   /* Term_Entry mmap */

    BACKET_ENTRY *m_be;				/* Hashtable pointer           */
    TERM_STR_ENTRY *m_te;			/* Term Entry pointer        */

    int m_termfile;					/* Term (text) file_no             */
    int m_term_entry_fd;			/* Term_Entry fd for huge files*/
};

#endif /* __PMINDEXH__ */
/*EOF*/
